# Errors and Exceptions

class TranslatorException(Exception):
    """Base exception to be able to catch all exceptions in this module"""
    pass


class ProviderAlreadyAdded(TranslatorException):
    def __init__(self, name):
        super().__init__(f"A provider with the name '{name}' has already been added.")


class LanguageNotSupported(TranslatorException):
    def __init__(self, language: str, direction: str = "source"):
        self.language = language
        self.direction = direction
        super().__init__(f"Language {language} is not supported with the current providers.")


class ProvidersMismatch(TranslatorException):
    def __init__(self, language):
        super().__init__(f"From language {language} is not supported with the chosen provider.")


class ProviderUnavailable(TranslatorException):
    def __init__(self, provider, to_language, source_language=None):
        super().__init__()
        self.provider = provider
        self.to_language = to_language
        self.source_language = source_language


class DetectedAsSameError(TranslatorException):
    def __init__(self, to_language, detected_language):
        super().__init__()
        self.to_language = to_language
        self.detected_language = detected_language


class NotEnoughCharacters(TranslatorException):
    def __init__(self, provider: str, characters: int, content_length: int, server_id: int):
        self.provider = provider.title()
        self.characters = characters
        self.content_length = content_length
        self.server_id = server_id

    def __str__(self):
        return "Unable to translate {} characters as the server {} does not have enough characters" \
               " for {}".format(self.content_length, self.server_id, self.provider)



from functools import cached_property
from types import MappingProxyType
from typing import Optional, Dict, Set, Mapping
from .caseinsensitivedict import CaseInsensitiveDict
from .abc import BaseProvider, Translation
from .errors import *


class AsyncTranslate:
    def __init__(self):
        self._providers: Dict[str, BaseProvider] = {}  # {'provider_name': ProviderInstance() }
        self._languages: Dict[str, Set[str]] = {}  # {'language_name': {'provider_name'} }
        self._language_names: Dict[str, str] = {}  # {'en': 'English'}

    @property
    def providers(self) -> Mapping[str, BaseProvider]:
        """Returns read-only copy of the providers"""
        return MappingProxyType(self._providers)

    @property
    def languages(self) -> Mapping[str, Set[str]]:
        """Returns read-only copy of the languages"""
        return MappingProxyType(self._languages)

    @property
    def language_names(self) -> Mapping[str, str]:
        return CaseInsensitiveDict(self._language_names)

    @cached_property
    def language_by_names(self) -> Mapping[str, str]:
        """Returns read-only copy of language codes with their English names"""
        return CaseInsensitiveDict({name: code for code, name in self._language_names.items()})

    @cached_property
    def default_provider(self) -> BaseProvider:
        """Returns the default (first) provider"""
        return next(iter(self._providers.values()))

    async def close(self):
        """Close all Provider aiohttp loops"""
        for provider in self._providers.values():
            await provider.close()

    async def add_provider(self, provider: BaseProvider):
        """Add a translator provider"""
        if not isinstance(provider, BaseProvider):
            raise TypeError('Providers must derive from BaseProvider')

        # Add to internal store
        provider_name = provider.name.casefold()
        if provider_name in self._providers:
            raise ProviderAlreadyAdded(provider_name)
        self._providers[provider_name] = provider

        languages = await provider.get_languages()
        for code, language_name in languages.items():
            self._language_names[code] = language_name
            if code in self._languages:
                self._languages[code].add(provider_name)
            else:
                self._languages[code] = {provider_name}

    async def add_providers(self, *backends: [BaseProvider]):
        """Add multiple providers"""
        for b in backends:
            await self.add_provider(b)

    def provider_for(self, language: str, preferred: Optional[str] = "") -> BaseProvider:
        try:
            available_providers = list(self._languages[language])
        except KeyError:
            raise LanguageNotSupported(language)
        return self._providers[preferred if preferred in available_providers else available_providers[0]]

    async def translate(self, to: str, content: str, provider: BaseProvider,
                        source_language: Optional[str] = None, **options) -> Translation:
        # Assumes to & source_language are valid language codes
        if source_language and source_language == to:
            raise DetectedAsSameError(to_language=to, detected_language=source_language)

        # Detect translating to/from same language
        detected_language = await self.providers.get('azure').detect(content.strip())
        if to == detected_language:
            raise DetectedAsSameError(to_language=to, detected_language=detected_language)
        return await provider.translate(content, to=to, source=source_language, **options)



import uuid
from json import JSONDecodeError
from typing import Dict, Optional, Union, Sequence, List, Any

from aiohttp_retry import RetryClient, ExponentialRetry

from async_translate.abc import BaseProvider, Translation
from async_translate.errors import TranslatorException, LanguageNotSupported
from async_translate.providers.azure.errors import NoAPIKeys, AllKeysExhausted, RequestException

MS_API_VER = "?api-version=3.0"
retry_options = ExponentialRetry(attempts=10, start_timeout=8.0, factor=1.3, statuses={429})


class Azure(BaseProvider):
    """
    Microsoft Azure Cognitive Services translator Backend
    """
    backend = "azure"
    ms_endpoint = "https://api.cognitive.microsofttranslator.com/"
    icon = "https://connectoricons-prod.azureedge.net/microsofttranslator/icon_1.0.1303.1871.png"

    def __init__(self, api_keys: Union[str, Sequence[str]]):
        if isinstance(api_keys, str):
            self._api_keys: List[str] = [api_keys]
        else:
            if len(api_keys) < 1:
                raise NoAPIKeys("No API keys provided")
            self._api_keys: Sequence[str] = api_keys
        self.api_iter = iter(self._api_keys)
        self.current_key = next(self.api_iter)
        self.session: Optional[RetryClient] = None

    @property
    def headers(self):
        return {
            'Ocp-Apim-Subscription-Key': self.current_key,
            'Content-type': 'application/json',
            'X-ClientTraceId': str(uuid.uuid4())
        }

    async def _request(self, endpoint, method='post', params: Optional[Dict] = None, json: Any = None,
                       api_key_override: Optional[str] = None):
        if params is None:
            params = {}

        headers = self.headers

        if api_key_override:
            headers['Ocp-Apim-Subscription-Key'] = api_key_override

        if accept_language := params.pop('accept_language', None):
            headers['Accept-Language'] = accept_language
        url = self.ms_endpoint + endpoint + MS_API_VER + "&"
        async with self.session.request(method=method, url=url, params=params, headers=headers, json=json) as resp:
            try:
                data = await resp.json(content_type=None)
                try:
                    if error := data.get('error'):
                        # Handle Free Quota Errors
                        if error['code'] in (401000, 403000, 403001, 429000, 429001, 429002):
                            try:
                                self.current_key = next(self.api_iter)
                            except StopIteration:
                                self.api_iter = iter(self._api_keys)
                                self.current_key = next(self.api_iter)
                                full_stop = True
                            else:
                                return await self._request(endpoint, method, params, json)
                            if full_stop:
                                raise AllKeysExhausted("All API Keys Exhausted", error)
                        else:  # Handle Generic Errors
                            raise RequestException(error, **error)
                except AttributeError:
                    pass
                return data
            except JSONDecodeError as e:
                raise TranslatorException(await resp.text()) from e

    async def get_languages(self, locale=None, *args, **kwargs) -> Dict[str, str]:
        # Setup retry client
        if not self.session:
            self.session = RetryClient(retry_options=retry_options)

        raw_languages = await self._request('languages', 'get',
                                            {'scope': 'translation', 'accept_language': locale})
        return {key: value['name'] for key, value in raw_languages['translation'].items()}

    async def close(self):
        # noinspection PyProtectedMember
        if self.session._closed:
            return
        await self.session.close()

    async def detect(self, content, key_override: Optional[str] = None) -> str:
        """Detect the language of the given content"""
        json_content = [{"text": content}]
        result = await self._request('detect', json=json_content, api_key_override=key_override)
        return result[0]['language']

    async def translate(self, content: str, to: str, source="", **options) -> Translation:
        """
        Translates text using the Azure API
        :param content: the text to translate
        :param to: the language code to translate to
        :param source: Optional language to translate from
        :return: the translation
        """
        params = {
            'to': to,
            'profanityAction': ('NoAction', 'Marked', 'Marked')[options.get('profanity_filter') or 0]
        }
        if source:
            params['from'] = source

        if params['profanityAction'] == 1:
            params['profanityMarker'] = 'Tag'

        json_content = [{"text": content}]
        try:
            translation_response = (await self._request('translate', params=params, json=json_content))
        except RequestException as te:
            if te.code in (400004, 400036, 400018, 400035):  # Invalid/missing language
                if te.code in (400018, 400035):  # Invalid 'from'
                    language = source
                    direction = "from"
                else:  # Invalid 'to'
                    language = to
                    direction = "to"
                raise LanguageNotSupported(language=language, direction=direction) from te
            raise

        entry = translation_response[0]
        return Translation(
            text=entry['translations'][0]['text'],
            to=entry['translations'][0]['to'],
            source=entry.get('detectedLanguage', {}).get('language')
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, Optional

ONE_DAY = 86400


@dataclass
class Translation:
    text: str
    to: str
    source: Optional[str] = None


class BaseProvider(ABC):
    icon = ""

    @property
    def name(self) -> str:
        """Shortcut to get name of the provider"""
        return self.__class__.__name__

    @abstractmethod
    async def get_languages(self, locale=None, *args, **kwargs) -> Dict[str, str]:
        raise NotImplementedError

    @abstractmethod
    async def detect(self, content) -> str:
        raise NotImplementedError

    @abstractmethod
    async def translate(self, content: str, to: str, source="", **options) -> Translation:
        raise NotImplementedError

    async def close(self):
        pass



import setuptools
import os
import versioneer

# Setup Optional requirements
extras_require = {}
suffix_len = len("_requirements.txt")
dirs = os.walk("async_translate/providers")
next(dirs)  # ignore the base folder

for provider_dir in dirs:
    name = provider_dir[0].split('/')[-1]
    dependencies = []
    try:
        with open(os.path.join("async_translate/providers", name,
                               "requirements.txt")) as fh:
            dependencies = fh.read().splitlines()
    except FileNotFoundError:
        continue
    extras_require[name] = dependencies

with open("README.md") as fh:
    long_description = fh.read()

setuptools.setup(
    name="async_translate",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="William Hatcher",
    author_email="william@memotic.net",
    description="Multi-provider async translate API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Memotic/async-translate",
    packages=setuptools.find_packages(),
    extras_require=extras_require,
    license='MIT',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Intended Audience :: Developers',
        'Natural Language :: English',
    ],
    python_requires='>=3.8',
)



from async_translate.errors import TranslatorException


class AllKeysExhausted(TranslatorException):
    pass


class RequestException(TranslatorException):
    def __init__(self, base, code: int, message: str):
        self.code: int = code
        self.message: str = message
        super().__init__(base)


class NoAPIKeys(TranslatorException):
    pass



import os
from typing import Dict

from aiocache import cached
from google.cloud.translate_v3.services.translation_service.async_client import TranslationServiceAsyncClient

from async_translate.abc import BaseProvider, ONE_DAY, Translation
from async_translate.errors import TranslatorException


class Google(BaseProvider):
    icon = "https://i.imgur.com/jDPXiQh.png"

    def __init__(self, credentials=None, parent: str = None):
        """
        :param credentials: Optional credentials dict
        :param parent: Optional string of parent project. If not present, will get from GOOGLE_TRANSLATE_PARENT env var
        """
        if credentials:
            self.client = TranslationServiceAsyncClient.from_service_account_info(credentials)
        else:
            self.client = TranslationServiceAsyncClient()
        try:
            self.parent = parent or os.environ['GOOGLE_TRANSLATE_PARENT']
        except KeyError:
            raise TranslatorException("Please set/export the 'GOOGLE_TRANSLATE_PARENT' environment variable. "
                                      "Ex: 'projects/mr-translate-1577912381600/locations/global'")

    @cached(ttl=ONE_DAY)
    async def get_languages(self, locale="en") -> Dict[str, str]:
        return {
            lang.language_code: lang.display_name
            for lang in filter(
                lambda l: l.support_source and l.support_target,
                (await self.client.get_supported_languages(parent=self.parent, display_language_code=locale)).languages)
        }

    async def detect(self, content) -> str:
        res = (await self.client.detect_language(parent=self.parent, content=content)).languages[0]
        return res.language_code

    async def translate(self, content: str, to: str, source="", **options) -> Translation:
        """
        Translates text using the Google API
        :param content: the text to translate
        :param to: the language code to translate to
        :param source: Optional language to translate from
        :return: the translation
        """
        params = {"contents": [content],
                  "target_language_code": to,
                  "parent": self.parent,
                  'mime_type': 'text/plain'}
        if source:
            params['source_language_code'] = source

        translation = (await self.client.translate_text(**params)).translations[0]
        return Translation(
            # text=unescape(translation.translated_text),
            text=translation.translated_text,
            to=to,
            source=translation.detected_language_code or None
        )
